/*
 * IEEE math operations in software
 * Copyright (c) 2006 Total Spectrum Software Inc.
 */

#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <fenv.h>

#define SINGLE /* single precision */
#ifdef TEST
#include <assert.h>
#define INCLUDE_SQRT
#endif

#define INTBITS 32
#define LONGLONGBITS 64

#ifdef SINGLE
// single precision
typedef int32_t FINT;
typedef uint32_t UFINT;
typedef uint16_t UHALFINT;
typedef float FTYPE;
#define COUNT_LEADING_ZEROS(x) __builtin_clz(x)

#define BITS 32
#define MANT_BITS 23
#define EXP_MASK (255)
#define EXP_OFF  (127)

#define FIXFSI __fixsfsi
#define FIXFUNSI __fixunssfsi
#define FIXFDI __fixsfdi
#define FIXFUNDI __fixunssfdi

#define FLOATUNDIF __floatundisf
#define FLOATUNSIF __floatunsisf

#define SQRT _sqrtf

#else
// double precision
typedef int64_t FINT;
typedef uint64_t UFINT;
typedef uint32_t UHALFINT;
typedef double FTYPE;
#define COUNT_LEADING_ZEROS(x) __builtin_clzll(x)

#define BITS 64
#define MANT_BITS 52
#define EXP_MASK (2047)
#define EXP_OFF  (1023)

#define FIXFSI __fixdfsi
#define FIXFUNSI __fixunsdfsi
#define FIXFDI __fixdfdi
#define FIXFUNDI __fixunsdfdi
#define SQRT _sqrtdf

#endif

#define SIGN_MASK ((UFINT)1<<(BITS-1))

#undef INF
#undef NAN
#define INF ((UFINT)EXP_MASK<<MANT_BITS)
#define NAN (INF|((UFINT)1<<(MANT_BITS-1))|(SIGN_MASK))

#define HALF SIGN_MASK
#define MANT_MASK (((UFINT)1<<MANT_BITS) - 1)
#define HIBITSET(x) (((FINT)x)<0)
#define HIBIT(x) ((UFINT)(x)<<(BITS-1))
#define FP_ONE ((UFINT)1<<MANT_BITS)
#define SHIFT_TO(s, bits) (((BITS-1)-((int)bits)) - ((int)s))

#define HIGHPART(u) ((UHALFINT)((u)>>(BITS/2)))
#define LOWPART(u)  ((UHALFINT)u)

typedef struct s_dblufint {
    UFINT hi;
    UFINT lo;
} DBLUFINT;

static DBLUFINT
add_double(DBLUFINT x, DBLUFINT y)
{
    x.lo = x.lo + y.lo;
    x.hi = x.hi + y.hi + (x.lo < y.lo);
    return x;
}

static DBLUFINT
umul_double(UFINT u, UFINT v)
{
    UHALFINT ul, vl, uh, vh;
    UFINT x0, x1, x2, x3;
    DBLUFINT r;

    ul = LOWPART(u); uh = HIGHPART(u);
    vl = LOWPART(v); vh = HIGHPART(v);

    x0 = (UFINT)ul * vl;
    x1 = (UFINT)ul * vh;
    x2 = (UFINT)uh * vl;
    x3 = (UFINT)uh * vh;

    x1 += HIGHPART(x0);
    x1 += x2;
    if (x1 < x2)
        x3 += ((UFINT)1<<(BITS/2));

    r.hi = x3 + HIGHPART(x1);
    r.lo = ((UFINT)LOWPART(x1) << (BITS/2)) + LOWPART(x0);
    return r;
}

/* calculate c = (a*b >> shift) */
static inline UFINT
umul_single(UFINT a, UFINT b, int shift)
{
    DBLUFINT r;
    UFINT c;

    r = umul_double(a, b);
    if (shift >= BITS) {
        shift -= BITS;
        c = r.hi;
        if (shift > 0)
            c = c >> shift;
    } else {
        c = r.hi << (BITS - shift);
        c |= r.lo >> shift;
    }
    return c;
}

uint32_t __round_bits_single = 0x80000000U;
uint32_t __round_mask_single = 0xFFFFFFFEU;

unsigned __fpflags = 0;

int __builtin_fesetround(int mode)
{
    switch(mode) {
    case FE_DOWNWARD:
        __round_bits_single = 1;
        __round_mask_single = 0;
        break;
    case FE_UPWARD:
        __round_bits_single = -1;
        __round_mask_single = 0;
        break;
    case FE_TOWARDZERO:
        __round_bits_single = 0;
        __round_mask_single = 0;
        break;
    case FE_TONEAREST:
        __round_bits_single = 0x80000000U;
        __round_mask_single = 0xFFFFFFFEU;
        break;
    default:
        return -1;
    }
    return 0;
}

#define RAISE(exc) __fpflags |= (exc)

/*
 * pack up a result
 * the high word of the mantissa is in a, the lo is in alo
 */
static UFINT
pack_result(UFINT a, UFINT alo, unsigned aexp, unsigned asign)
{
    int shift;
    UFINT rounding;

    /* re-normalize */
    if (a == 0) {
        if (alo != 0) {
            aexp -= MANT_BITS;
            shift = COUNT_LEADING_ZEROS(alo);
            aexp -= shift;
            shift = SHIFT_TO(shift, MANT_BITS);
            if (shift >= 0) {
                a = alo << (shift);
            } else {
                shift = -shift;
                a = alo >> shift;
            }
            alo = 1;
        } else {
            aexp = 0;
        }
    } else {
        shift = COUNT_LEADING_ZEROS(a);
        shift = SHIFT_TO(shift, MANT_BITS);
        aexp += shift;
        if (shift > 0) {
            alo = (alo>>shift) | (alo != 0);
            alo = alo | (a<<(BITS-shift));
            a = a>>shift;
        } else if (shift < 0) {
            shift = -shift;
            a = a<<shift;
            a = a | (alo >> (BITS-shift));
            alo = alo << shift;
        }
    }
    /* check for overflow or denormalized numbers */
    if (aexp >= EXP_MASK) {
        RAISE(FE_OVERFLOW);
        aexp = EXP_MASK-1;
        a = FP_ONE - 1;
        alo = ((UFINT)1<<(BITS-1)) | 1;  /* ready to be rounded up; also will force INEXACT */
    } else if (aexp <= 0) {
        /* possible underflow */
        shift = (-aexp+1);
        if (shift < MANT_BITS) {
            alo = (alo >> shift) | (alo != 0);
            alo = alo | (a<<(BITS-shift));
            a = a>>shift;
        } else {
            a = 0;
            alo = 1;
        }
        aexp = 0;
    } else {
        a -= FP_ONE;
    }

    /* now round */
    if (alo != 0) {
        RAISE(FE_INEXACT);
        if (aexp == 0)
            RAISE(FE_UNDERFLOW);
    }
    rounding = __round_bits_float;
    if (asign)
        rounding = -rounding;
    alo += rounding;
    if (alo < rounding) /* carry */
        a += 1;
    if (alo == 0) {
        a &= __round_mask_float;
    }
    a += ((UFINT)aexp << MANT_BITS);
    a |= HIBIT(asign);
    return a;
}

UFINT
_float_add(UFINT a, UFINT b)
{
    unsigned aexp, bexp;
    UFINT amag, bmag;
    unsigned asign;
    unsigned eorsign;
    int shift;
    UFINT alo;

    amag = a+a;
    bmag = b+b;
    if (amag < bmag) {
        UFINT temp;
        temp = a;
        a = b;
        b = temp;
    }
    asign = a >> (BITS-1);
    eorsign = asign ^ (b>>(BITS-1));
    aexp = (a >> MANT_BITS) & EXP_MASK;
    bexp = (b >> MANT_BITS) & EXP_MASK;
    a &= MANT_MASK;
    b &= MANT_MASK;

    if (bexp == 0) {
        if (b == 0) {
            /* b is 0; normally this means we want to return a,
               except in the case that a==-0 and b==0
            */
            if (aexp == 0 && a == 0 && asign != 0) {
                /* a is -0; return +0 if b is +0, i.e. */
                if (eorsign)
                    asign = 0;
            }
            goto return_a;
        } else {
            /* denormalized number */
            b = b<<1;
        }
    } else {
        b |= FP_ONE;
    }

    if (aexp == 0) {
        /* a is 0 only if b is 0, and we've already handled that */
        a = a<<1;
    } else if (aexp == EXP_MASK) {
        /* we know a is the larger magnitude, so b is NaN only if a is */
        if (a == 0) {
            /* both are infinities */
            if (eorsign) {
                /* inf - inf == nan */
                RAISE(FE_INVALID);
                return HIBIT(asign) | NAN;
            }
            RAISE(FE_INEXACT);
            goto return_a;
        } else {
            /* a is already a NaN, just return it */
            goto return_a;
        }
    } else {
        a |= FP_ONE;
    }

    if (eorsign) {
        b = -b;
    }
    shift = aexp - bexp;
    if (shift > 0) {
        if (shift >= BITS) {
            /* make sure it's a signed right shift */
            b = ((FINT)b)>>(BITS-1);
            alo = b|1;
        } else {
            alo = (b<<(BITS-shift));
            b = ((FINT)b)>>shift;
        }
    } else {
        alo = 0;
    }
    a = a+b;

    return pack_result(a, alo, aexp, asign);

return_a:
    a += ((UFINT)aexp << MANT_BITS);
    a |= HIBIT(asign);
    return a;
}

UFINT
_float_sub(UFINT a, UFINT b)
{
    return FN3(add)(a, b^SIGN_MASK);
}

UFINT
_float_mul(UFINT a, UFINT b)
{
    unsigned aexp, bexp;
    unsigned origasign;
    unsigned asign;
    int shift;
    DBLUFINT r;

    origasign = a >> (BITS-1);
    asign = origasign ^ (b>>(BITS-1));
    aexp = (a >> MANT_BITS) & EXP_MASK;
    bexp = (b >> MANT_BITS) & EXP_MASK;
    a &= MANT_MASK;
    b &= MANT_MASK;

    if (bexp == 0) {
        if (b == 0) {
            /* b is 0; normally this means we want to return 0,
               except in the case that a==NaN or INF
            */
            if (aexp == EXP_MASK) {
                if (a != 0) {
                    /* return a */
                    asign = origasign;
                    goto return_a;
                }
                /* otherwise 0 * infinity = NAN */
            invalid_mult:
                RAISE(FE_INVALID);
                return HIBIT(asign) | NAN;
            }
            return HIBIT(asign);
        } else {
            /* denormalized number */
            shift = COUNT_LEADING_ZEROS(b) - MANT_BITS;
            b = b<<shift;
            bexp -= (shift-1);
        }
    } else if (bexp == EXP_MASK) {
        /* b is infinity or NAN */
        if (b != 0) {
            /* b is NaN -- return it */
            a = b;
            aexp = bexp;
            goto return_a;
        }
        if (aexp == EXP_MASK) {
            /* b is infinity, so is a; return a */
            RAISE(FE_INEXACT);
            goto return_a;
        } else if (aexp == 0 && a == 0) {
            /* infinity * 0: invalid */
            goto invalid_mult;
        }
    } else {
        b |= FP_ONE;
    }

    if (aexp == 0) {
        if (a == 0) {
            /* a is 0; normally this means we want to return 0,
               except in the case that b==NaN or INF;
               but we handled those already
            */
            return HIBIT(asign);
        } else {
            /* denormalized number */
            shift = COUNT_LEADING_ZEROS(a) - MANT_BITS;
            a = a<<shift;
            aexp -= (shift-1);
        }
    } else if (aexp == EXP_MASK) {
        /* we already checked b for NaN or infinity or 0 */
        if (a == 0) {
            /* infinity */
            RAISE(FE_INEXACT);
            goto return_a;
        } else {
            /* a is already a NaN, just return it */
            goto return_a;
        }
    } else {
        a |= FP_ONE;
    }


    r = umul_double(a, b);
    aexp = aexp + bexp - EXP_OFF;
    aexp += (BITS - MANT_BITS);  // correct for umul_double effects

    return pack_result(r.hi, r.lo, aexp, asign);

return_a:
    a += ((UFINT)aexp << MANT_BITS);
    a |= HIBIT(asign);
    return a;
}

UFINT
_float_div(UFINT a, UFINT b)
{
    unsigned aexp, bexp;
    unsigned origasign;
    unsigned asign;
    int shift;
    int bits;
    DBLUFINT r;
    UFINT q;
    UFINT origa, origb;

    origasign = a >> (BITS-1);
    asign = origasign ^ (b>>(BITS-1));
    aexp = (a >> MANT_BITS) & EXP_MASK;
    bexp = (b >> MANT_BITS) & EXP_MASK;
    a &= MANT_MASK;
    b &= MANT_MASK;

    if (bexp == 0) {
        if (b == 0) {
            /* b is 0; this is a divide by 0
               except in the case that a==NaN
            */
            if (aexp == EXP_MASK) {
                if (a != 0) {
                    /* return a */
                    asign = origasign;
                    goto return_a;
                }
            } else {
                RAISE(FE_DIVBYZERO);
            }
            /* special case 0/0 */
            if (aexp == 0 && a == 0) {
                RAISE(FE_INVALID);
                return HIBIT(asign) | NAN;
            }
            RAISE(FE_INEXACT);
            return HIBIT(asign) | INF;
        } else {
            /* denormalized number */
            shift = COUNT_LEADING_ZEROS(b) - MANT_BITS;
            b = b<<shift;
            bexp -= (shift-1);
        }
    } else if (bexp == EXP_MASK) {
        /* b is infinity or NAN */
        if (b != 0) {
            /* b is NaN -- return it */
            a = b;
            aexp = bexp;
            origasign = asign ^ origasign;
            goto return_a;
        }
        if (aexp == EXP_MASK) {
            if (a == 0) {
                /* b is infinity, so is a; return NAN */
                RAISE(FE_INVALID);
                return HIBIT(asign) | NAN;
            } else {
                /* a is NaN -- return it */
                goto return_a;
            }
        }
        RAISE(FE_INEXACT);
        return HIBIT(asign);
    } else {
        b |= FP_ONE;
    }

    if (aexp == 0) {
        if (a == 0) {
            /* a is 0; normally this means we want to return 0,
               except in the case that b==NaN or INF;
               but we handled those already
            */
            return HIBIT(asign);
        } else {
            /* denormalized number */
            shift = COUNT_LEADING_ZEROS(a) - MANT_BITS;
            a = a<<shift;
            aexp -= (shift-1);
        }
    } else if (aexp == EXP_MASK) {
        /* we already checked b for NaN or infinity or 0 */
        if (a == 0) {
            /* infinity */
            RAISE(FE_INEXACT);
            goto return_a;
        } else {
            /* a is already a NaN, just return it */
            goto return_a;
        }
    } else {
        a |= FP_ONE;
    }

    aexp = (aexp - bexp) + EXP_OFF;

#define DIV_EXTRA_BITS 4
#define DIV_BITS (MANT_BITS+DIV_EXTRA_BITS)
#define DIV_ADJ (BITS - (DIV_BITS))

    origa = a;
    origb = b;
    a = a << (DIV_BITS-MANT_BITS);
    b = b << (DIV_BITS-MANT_BITS);
    /* make q an approximation of 1/b, good to about 4 bits,
       and with MANTBITS+1 bits of accuracy
    */
    q = (UFINT)(1.457106781 * ((UFINT)1<<(DIV_BITS)));
    q = q - (b>>1);

    /* now a/b = q*a / q*b */
    bits = 4;
    do {
#if 0
        r = umul_double(q, b);
        b = (r.hi << DIV_ADJ) | (r.lo >> (BITS-DIV_ADJ));
        r = umul_double(q, a);
        a = (r.hi << DIV_ADJ) | (r.lo >> (BITS-DIV_ADJ));
#else
        b = umul_single(q, b, DIV_BITS);
        a = umul_single(q, a, DIV_BITS);
#endif
        q = ((UFINT)2<<(DIV_BITS)) - b;
        bits += bits;
    } while (bits <= BITS);

    /* at this point a is approximately origa/origb */
    /* see how good the approximation is */

    /* first, try rounding a up */
    a = (a+1) & ~1;

    /* make sure a*origb < origa */
    r = umul_double(origb, a);
    r.hi = (r.hi << DIV_ADJ) | (r.lo >> (BITS-DIV_ADJ));
    r.lo = (r.lo << DIV_ADJ);
    while (r.hi > origa || (r.hi == origa && r.lo > 0)) {
        --a;
        r = umul_double(origb, a);
        r.hi = (r.hi << DIV_ADJ) | (r.lo >> (BITS-DIV_ADJ));
        r.lo = (r.lo << DIV_ADJ);
    }
    /* if the result isn't exact, flag that */
    if (r.hi < origa) {
        a |= 1;  /* make sure bottom bits of quotient are nonzero */
    }

    return pack_result(a, 0, aexp-DIV_EXTRA_BITS, asign);
return_a:
    a += ((UFINT)aexp << MANT_BITS);
    a |= HIBIT(origasign);
    return a;
}

static inline int
_float_cmp(FINT a, FINT b)
{
    int asign, bsign;
    unsigned aexp, bexp;
    FINT delta;
    int unordered = 0x80000000U;
    /* check first for NaNs */
    aexp = (a >> MANT_BITS) & EXP_MASK;
    if (aexp == EXP_MASK) {
        if ((a & MANT_MASK) != 0)
            return unordered;
    }
    bexp = (b >> MANT_BITS) & EXP_MASK;
    if (bexp == EXP_MASK) {
        if ((b & MANT_MASK) != 0)
            return unordered;
    }

    asign = a >> (BITS-1);
    bsign = b >> (BITS-1);

    if ( ((asign ^ bsign) != 0) ) {
        /* if they are both 0's, return 0 */
        if ((a|b) == SIGN_MASK)
            return 0;
        /* otherwise, return sign of a */
        return (asign | 1);
    }
    /* same signs; calculate delta, return sign of it */
    if (asign) {
        delta = b-a;
    } else {
        delta = a-b;
    }
    if (delta == 0)
        return 0;
    bsign = delta >> (BITS-1);
    return (bsign | 1);
}

int FN2(eq)(FINT a, FINT b) { return cmpf(a, b, -1); }
int FN2(ne)(FINT a, FINT b) { return cmpf(a, b, -1); }
int FN2(gt)(FINT a, FINT b) { return cmpf(a, b, -1); }
int FN2(ge)(FINT a, FINT b) { return cmpf(a, b, -1); }
int FN2(lt)(FINT a, FINT b) { return cmpf(a, b, 1); }
int FN2(le)(FINT a, FINT b) { return cmpf(a, b, 1); }

UFINT
FN0(floatunsi)(unsigned int x)
{
    return FN0(floatundi)((unsigned long long)x);
}

UFINT
FN0(floatdi)(long long x)
{
    UFINT a;
    UFINT sign = 0;
    if (x < 0) {
        sign = ((UFINT)1<<(BITS-1));
        x = -x;
    }
    a = FN0(floatundi)(x);
    a |= sign;
    return a;
}

UFINT
FN0(floatsi)(int x)
{
    return FN0(floatdi)((long long)x);
}


/*
 * conversion from float to unsigned or signed long long
 * maxbits gives maximum bits to use
 * is_signed is 1 for signed numbers, 0 for unsigned
 */
static unsigned long long
fixconvert(UFINT a, int is_signed, int maxbits)
{
    unsigned asign;
    unsigned aexp;
    int shift;
    UFINT alo;
    unsigned long long x;

    maxbits -= is_signed;
    asign = a >> (BITS-1);
    aexp = (a>>MANT_BITS) & EXP_MASK;
    if (aexp < EXP_OFF || (asign && !is_signed) ) {
        /* underflow, unless a was 0 */
        if (a != 0) {
            RAISE(FE_UNDERFLOW);
        }
        return 0;
    }
    a &= MANT_MASK;
    if (aexp == EXP_MASK) {
        /* infinity, or NaN */
        if (a != 0) {
            RAISE(FE_INVALID);
        } else
    overflow:
        RAISE(FE_OVERFLOW);
        return is_signed ? (1LL<<maxbits) : 0xFFFFFFFFFFFFFFFFLL;
    }
    aexp -= EXP_OFF;
    if (aexp > maxbits) {
        goto overflow;
    }
    a |= FP_ONE;
    shift =  MANT_BITS - aexp;
    if (shift >= 0) {
        alo = (a<<(BITS-shift));
        x = a >> shift;
        if (alo != 0) {
            RAISE(FE_INEXACT);
        }
        if (asign)
            x = -x;
        return x;
    }
    shift = -shift;
    x = a << shift;
    if (asign)
        x = -x;
    return x;
}

long long
FIXFDI(UFINT a)
{
  return fixconvert(a, 1, LONGLONGBITS);
}

unsigned long long
FIXFUNDI(UFINT a)
{
  return fixconvert(a, 0, LONGLONGBITS);
}

int
FIXFSI(UFINT a)
{
  return fixconvert(a, 1, INTBITS);
}

unsigned int
FIXFUNSI(UFINT a)
{
  return fixconvert(a, 0, INTBITS);
}


#if defined(INCLUDE_SQRT)
/*
 * square root
 */
UFINT
SQRT(UFINT a)
{
    int aexp;
    int shift;
    unsigned asign;
    UFINT origa;
    UFINT g, h, r, d;
    DBLUFINT gdbl;
    int i;
    int maxiter;

    /* check for 0 or -0 */
    if ((-a) == a)
        return 0;
    origa = a;
    aexp = (a >> MANT_BITS) & EXP_MASK;
    asign = (a >> (BITS-1));
    a &= MANT_MASK;

    if (aexp == EXP_MASK) {
        if (a != 0) {
            /* a NaN */
            return origa;
        }
        /* a is infinite */
        if (asign) {
            RAISE(FE_INVALID);
            return NAN;
        }
        RAISE(FE_INEXACT);
        return origa;
    }
    /* sqrt(negative number) */
    if (asign) {
        RAISE(FE_INVALID);
        return NAN;
    }
    if (aexp == 0) {
        /* zero or denormalized number
           we already checked for zero above, so it must
           be a denormalized number
        */
        int shift;
        shift = COUNT_LEADING_ZEROS(a) - MANT_BITS;
        a = a<<shift;
        aexp -= (shift-1);
    } else {
        a |= ((UFINT)1<<MANT_BITS);
    }

    /* a note on initial estimates;
       for convergence; if we estimate r0=approximate(1/sqrt(a))
       then we want 1/2 < a*r0*r0 < 3/2 in order to converge
       for 1 <= a < 2, use r0 = (5 - a)/4 (so r0(1) = 1, r0(2) = 3/4
           2 <= a < 4, use r0 = 1 - a/8 (so r0(2) = 3/4, r0(4) = 1/2
       in both cases the worst performance is at r0(2) = 3/4, and we have:
             a*r0*r0 = 2*(3/4)*(3/4) = 18/16

    */
    aexp = (aexp - EXP_OFF);
    /* add some extra precision to a */
    a = a << DIV_EXTRA_BITS;

    if (aexp & 1) {
        /* odd exponent, so number is >= 2.0 and < 4 */
        a = (a<<1);
        r = (((UFINT)1<<DIV_BITS)) - (a >> 3);
    } else {
        r = (((UFINT)5<<DIV_BITS) - a) >> 2;
    }
    aexp = aexp >> 1;

    /* OK, at this point r is 8/sqrt(a), to about 4 bits */
    g = umul_single(a, r, DIV_BITS);
    h = r>>1;
    /* g is a guess for sqrt(a) accurate to about 4 bits */


#ifdef SINGLE
    maxiter = 3;
#else
    maxiter = 5;
#endif

    for (i = 0; i < maxiter; i++) {
        r = ((UFINT)3<<(DIV_BITS-1)) - umul_single(h, g, DIV_BITS);
        g = umul_single(g, r, DIV_BITS);
        h = umul_single(h, r, DIV_BITS);
        /* g is a guess for sqrt(a) to about 8 bits */
    }

    /* final iteration, full result */
    r = ((UFINT)3<<(DIV_BITS-1)) - umul_single(h, g, DIV_BITS);
    gdbl = umul_double(g, r);
    shift = DIV_BITS + DIV_EXTRA_BITS;
    if (shift >= BITS) {
        shift -= BITS;
        gdbl.lo >>= shift;
        gdbl.lo |= gdbl.hi << (BITS - shift);
        gdbl.hi >>= shift;
    } else {
        shift = BITS - shift;
        gdbl.hi <<= shift;
        gdbl.hi |= gdbl.lo >> (BITS - shift);
        gdbl.lo <<= shift;
    }
    return pack_result(gdbl.hi, gdbl.lo, aexp + EXP_OFF, asign);
}
#endif

#ifdef TEST
static void
validateadd(FTYPE a, FTYPE b)
{
    union fi { FTYPE f; UFINT i; } ua, ub, uc;
    UFINT c;

    ua.f = a;
    ub.f = b;
    uc.f = a+b;
    c = FN3(add)(ua.i, ub.i);
    if (c != uc.i) {
        printf("%16llx + %16llx -> %16llx (should be %16llx)\n",
               (long long)ua.i, (long long)ub.i, (long long)c, (long long)uc.i);
    }
    c = FN3(add)(ub.i, ua.i);
    if (c != uc.i) {
        printf("%16llx + %16llx -> %16llx (should be %16llx)\n",
               (long long)ub.i, (long long)ua.i, (long long)c, (long long)uc.i);
    }
}

static void
validatemul(FTYPE a, FTYPE b)
{
    union fi { FTYPE f; UFINT i; } ua, ub, uc;
    UFINT c;

    ua.f = a;
    ub.f = b;
    uc.f = a*b;
    c = FN3(mul)(ua.i, ub.i);
    if (c != uc.i) {
        printf("%16llx * %16llx -> %16llx (should be %16llx)\n",
               (long long)ua.i, (long long)ub.i, (long long)c, (long long)uc.i);
    }
    c = FN3(mul)(ub.i, ua.i);
    if (c != uc.i) {
        printf("%16llx * %16llx -> %16llx (should be %16llx)\n",
               (long long)ub.i, (long long)ua.i, (long long)c, (long long)uc.i);
    }
}

static void
validatediv(FTYPE a, FTYPE b)
{
    union fi { FTYPE f; UFINT i; } ua, ub, uc;
    UFINT c;

    ua.f = a;
    ub.f = b;
    uc.f = a/b;
    c = FN3(div)(ua.i, ub.i);
    if (c != uc.i) {
        printf("%16llx / %16llx -> %16llx (should be %16llx)\n",
               (long long)ua.i, (long long)ub.i, (long long)c, (long long)uc.i);
    }
}

static void
validatesqrt(FTYPE a)
{
    union fi { FTYPE f; UFINT i; } ua, ub, uc;
    UFINT c;

    ua.f = a;
#ifdef SINGLE
    uc.f = sqrtf(a);
#else
    uc.f = sqrt(a);
#endif
    c = SQRT(ua.i);
    if (c != uc.i) {
        printf("sqrt(%f=%16llx) -> %16llx (should be %16llx)\n",
               (double)ua.f, (long long)ua.i, (long long)c, (long long)uc.i);
    }
}

void
validatetofloat(int x)
{
    union fi { FTYPE f; UFINT i; } uc;
    UFINT c;

    uc.f = (FTYPE)x;
    c = FN0(floatsi)(x);
    if (c != uc.i) {
        printf("%x -> %16llx (should be %16llx)\n",
               (int)x, (long long)c, (long long)uc.i);
    }
}

void
validatefromfloat(FTYPE a)
{
    union fi { FTYPE f; UFINT i; } ua;
    int xo, xn;

    ua.f = a;
    xo = (int)a;
    xn = FIXFSI(ua.i);

    if (xo != xn) {
        printf("%16llx -> %x (should be %x)\n",
               (long long)ua.i, xn, xo);
    }
}

void
validatecmp(FTYPE a, FTYPE b)
{
    int i0, i1, r;
    union fi {FTYPE f; UFINT i; } ua, ub;

    ua.f = a; ub.f = b;

    i0 = a < b;
    i1 = (r = FN2(lt)(ua.i, ub.i)) < 0;
    assert(i0 == i1);
    i0 = a <= b;
    i1 = (r = FN2(le)(ua.i, ub.i)) <= 0;
    assert(i0 == i1);
    i0 = a > b;
    i1 = (r = FN2(gt)(ua.i, ub.i)) > 0;
    assert(i0 == i1);
    i0 = a >= b;
    i1 = (r = FN2(ge)(ua.i, ub.i)) >= 0;
    assert(i0 == i1);
    i0 = a == b;
    i1 = (r = FN2(eq)(ua.i, ub.i)) == 0;
    assert(i0 == i1);
    i0 = a != b;
    i1 = (r = FN2(ne)(ua.i, ub.i)) != 0;
    assert(i0 == i1);

}

#undef INF
#undef NAN

#define INF __builtin_inf()
#define NAN __builtin_nan("")

int
main()
{
    validatecmp(-3.0, -2.0);
    validatecmp(0.0, -0.0);
    validatecmp(1.0, 1.0);
    validatecmp(-1.0, 2.0);
    validatecmp(2.1, 0.1);
    validatecmp(2.1, INF);
    validatecmp(2.1, -INF);
    validatecmp(NAN, 0.0);
    validatecmp(NAN, NAN);

    validateadd(100.0,  1.0);
    validateadd(100.0, -1.0);
    validateadd(1.0, -0.5);
    validateadd(1.0, 0.5);
    validateadd(1.0, 1.0);
    validateadd(1.0, 2.0);
    validateadd(1.0, 3.0);
    validateadd(1.1, 0.9);
    validateadd(0.0, 0.0);
    validateadd(1.0, INF);
    validateadd(-0.0, 0.0);
    validateadd(-0.0, -0.0);
    validateadd(INF, 1.0);
    validateadd(INF, -INF);
    validateadd(NAN, 2.0);

    validatemul(1.0,  1.0);
    validatemul(2.0,  2.0);
    validatemul(2.0,  -2.0);
    validatemul(100.0, -0.1);
    validatemul(-0.0, 0.0);
    validateadd(2.0, INF);
    validateadd(0.0, INF);

    validatediv(1.0, 1.0);
    validatediv(-1.0, 2.0);
    validatediv(100, 10.0);
    validatediv(0.0,0.0);
    validatediv(1.0,0.0);
    validatediv(INF, 2.0);
    validatediv(1.0,-INF);
    validatediv(NAN,-INF);
    validatediv(NAN,-1.0);

    validatetofloat(0);
    validatetofloat(1);
    validatetofloat(2);
    validatetofloat(0x80000000);
    validatetofloat(0x7fffffff);

    validatefromfloat(0.0);
    validatefromfloat(-0.0);
    validatefromfloat(1.0);
    validatefromfloat(-4.0);
    validatefromfloat((double)0x80000000);
    validatefromfloat((double)0x7fffffff);
    validatefromfloat((double)-0x7ffffff);

    validatesqrt(0.0);
    validatesqrt(1.0);
    validatesqrt(2.0);
    validatesqrt(2.9);
    validatesqrt(3.0);
    validatesqrt(3.15);
    validatesqrt(4.0);

    return 0;
}
#endif
